import random

from math import ceil, sqrt

import numpy as np


class Bytes(np.ndarray):
    """
    numpy.ndarray which dtype is uint8 and only have one axis.
    """
    def __new__(cls):
        return np.array([], dtype=np.uint8).astype(np.uint8).view(Bytes)

    def __getitem__(self, key):
        return super(Bytes, self).__getitem__(key)

    def __add__(self, b):
        return np.append(self, b).astype(np.uint8).view(Bytes)

    @classmethod
    def from_str(cls, str: str, encoding='utf-8'):
        return np.frombuffer(str.encode(encoding), dtype=np.uint8).view(cls)

    def to_str(self, encoding='utf-8') -> str:
        return self.tobytes().decode(encoding)

    @classmethod
    def from_ndarray(cls, array: np.ndarray):
        return np.reshape(array, (-1)).astype(np.uint8).view(Bytes)

    def to_ndarray(self, shape: tuple, dtype=np.uint8) -> np.ndarray:
        return np.resize(self, shape).astype(dtype)


class Utils:
    @classmethod
    def image2bytes(cls, image: np.ndarray) -> Bytes:
        """
        Wrapper for Bytes.from_ndarray. Provides a unified interface.
        """
        return Bytes.from_ndarray(image)

    @classmethod
    def bytes2image(cls, data: Bytes) -> np.ndarray:
        """
        Converts a Bytes object into a square image with a minimum side length of 100.
        """
        size = data.size
        sidelen = max(100, ceil(sqrt(ceil(size/3))))
        return data.to_ndarray((sidelen, sidelen, 3))

    @classmethod
    def randstr(cls, len: int, charset='abcdefghijklmnopqrstuvwxyz0123456789') -> str:
        return ''.join(random.sample(charset, len))
